#include "P2PTunnelAgent.h"
#include "sample_constants.h"
#include <pthread.h>
#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#define ENABLE_TUNNEL_SERVICE 0

#define MAX_SERVER_CONNECT_NUM 4

#define HTTP_MAPPING_LOCAL_PORT 10000
#define HTTP_MAPPING_REMOTE_PORT 80

#define HTTPS_MAPPING_LOCAL_PORT 15000

#define SSH_MAPPING_LOCAL_PORT 20000
#define SSH_MAPPING_REMOTE_PORT 22

#define TELNET_MAPPING_LOCAL_PORT 30000
#define TELNET_MAPPING_REMOTE_PORT 23

#define TUNNEL_USERNAME "admin"
#define TUNNEL_PASSWORD "testpwd"

typedef struct st_AuthData {
    char szUsername[64];
    char szPassword[64];
} sAuthData;

typedef struct st_TunnelInfo {
    char uid[21];
    int nHttpIndex;
    int nHttpsIndex;
    int nTelnetIndex;
    int nSshIndex;
    int nRetryConnectFailedCnt;
    int nSID;
    bool bException;
} sTunnelInfo;

//########################################################
//# P2PTunnel Agent
//########################################################
static sTunnelInfo gsTunnelInfo[MAX_SERVER_CONNECT_NUM];

static void tunnelClientAuthentication(char *cszAccount, uint32_t nAccountMaxLength, char *cszPassword, uint32_t nPasswordMaxLength, const void *pArg) {
    sTunnelInfo *info = (sTunnelInfo *)pArg;
    printf("[%s] nAccountMaxLength=%d, nPasswordMaxLength=%d, UID=%s\n", __func__, nAccountMaxLength, nPasswordMaxLength, info->uid);
    strcpy(cszAccount, TUNNEL_USERNAME);
    strcpy(cszPassword, TUNNEL_PASSWORD);
}

static void tunnelStatusHandle(int nErrorCode, int nSID, void *pArg) {
    if (nErrorCode == TUNNEL_ER_DISCONNECTED) {
        printf("SID[%d] TUNNEL_ER_DISCONNECTED Log file here!\n", nSID);
        for (int i = 0; i < MAX_SERVER_CONNECT_NUM; i++) {
            if (gsTunnelInfo[i].nSID == nSID) {
                gsTunnelInfo[i].bException = true;
                break;
            }
        }
        if (pArg != NULL) {
            printf("pArg = %s\n", (char *)pArg);
        }
    }
}

static void tunnelExceptionCheck(void) {
    for (int i = 0; i < MAX_SERVER_CONNECT_NUM; i++) {
        if (gsTunnelInfo[i].bException) {
            int nSID = gsTunnelInfo[i].nSID;

            P2PTunnelAgent_StopPortMapping(gsTunnelInfo[i].nHttpIndex);
#if ENABLE_TUNNEL_SERVICE
            P2PTunnelAgent_StopServiceMapping(gsTunnelInfo[i].nHttpsIndex);
#endif
            P2PTunnelAgent_StopPortMapping(gsTunnelInfo[i].nTelnetIndex);
            P2PTunnelAgent_StopPortMapping(gsTunnelInfo[i].nSshIndex);
            P2PTunnelAgent_Disconnect(nSID);

            gsTunnelInfo[i].nSID = -1;
            gsTunnelInfo[i].bException = false;
        }
    }
}

static void tunnelAgentConnect(sTunnelInfo *info, int index) {
    int ret = P2PTunnelAgent_Connect_Ex(info->uid, tunnelClientAuthentication, info);
    if (ret == TUNNEL_ER_REMOTE_NOT_SUPPORT_DTLS) {
        printf("Remote P2PTunnel is old version, try P2PTunnelAgent_Connect()\n");
        sAuthData authData = {.szUsername = {0}, .szPassword = {0}};
        int nErrFromDeviceCB = 0;
        strcpy(authData.szUsername, "Tutk.com");
        strcpy(authData.szPassword, "P2P Platform");
        
        ret = P2PTunnelAgent_Connect(info->uid, (void *)&authData, sizeof(sAuthData), &nErrFromDeviceCB);
        printf("P2PTunnelAgent_Connect UID[%s] nErrFromDeviceCB[%d]\n", info->uid, nErrFromDeviceCB);
    }
    if (ret < 0) {
        printf("UID[%s] tunnel Connect error[%d]\n", info->uid, ret);
        return;
    } else {
        info->nSID = ret;
        printf("UID[%s] tunnel connect success\n", info->uid);
    }

    ret = P2PTunnel_SetBufSize(info->nSID, 5120000);
    if (ret != TUNNEL_ER_NoERROR) {
        printf("P2PTunnel_SetBufSize() UID[%s] error[%d]\n", info->uid, ret);
    }

    info->nHttpIndex = P2PTunnelAgent_PortMapping(info->nSID, HTTP_MAPPING_LOCAL_PORT + index, HTTP_MAPPING_REMOTE_PORT);
    if (info->nHttpIndex < 0) {
        printf("P2PTunnelAgent_PortMapping() HTTP UID[%s] error[%d]!\n", info->uid, info->nHttpIndex);
    } else {
        printf("UID[%s] HTTP mapping success, LOCAL[%d]<->REMOTE[%d]\n", info->uid, HTTP_MAPPING_LOCAL_PORT + index, HTTP_MAPPING_REMOTE_PORT);
    }

#if ENABLE_TUNNEL_SERVICE
    info->nHttpsIndex = P2PTunnelAgent_ServiceMapping(info->nSID, HTTPS_MAPPING_LOCAL_PORT + index, "https");
    if (info->nHttpsIndex < 0) {
        printf("P2PTunnelAgent_ServiceMapping() HTTPS UID[%s] error[%d]!\n", info->uid, info->nHttpsIndex);
    } else {
        printf("UID[%s] HTTPS mapping success, LOCAL[%d]<->REMOTE[https]\n", info->uid, HTTPS_MAPPING_LOCAL_PORT + index);
    }
#endif

    info->nTelnetIndex = P2PTunnelAgent_PortMapping(info->nSID, TELNET_MAPPING_LOCAL_PORT + index, TELNET_MAPPING_REMOTE_PORT);
    if (info->nTelnetIndex < 0) {
        printf("P2PTunnelAgent_PortMapping() TELENT UID[%s] error[%d]!\n", info->uid, info->nTelnetIndex);
    } else {
        printf("UID[%s] TELENT mapping success, LOCAL[%d]<->REMOTE[%d]\n", info->uid, TELNET_MAPPING_LOCAL_PORT + index, TELNET_MAPPING_REMOTE_PORT);
    }

    info->nSshIndex = P2PTunnelAgent_PortMapping(info->nSID, SSH_MAPPING_LOCAL_PORT + index, SSH_MAPPING_REMOTE_PORT);
    if (info->nSshIndex < 0) {
        printf("P2PTunnelAgent_PortMapping() SSH UID[%s] error[%d]!\n", info->uid, info->nSshIndex);
    } else {
        printf("UID[%s] SSH mapping success, LOCAL[%d]<->REMOTE[%d]\n", info->uid, SSH_MAPPING_LOCAL_PORT + index, SSH_MAPPING_REMOTE_PORT);
    }

    info->nRetryConnectFailedCnt = 0;

    return;
}

static void tunnelAgentDisconnectAll() {
    for (int i = 0; i < MAX_SERVER_CONNECT_NUM; i++) {
        if (gsTunnelInfo[i].nSID >= 0) {
            P2PTunnelAgent_Disconnect(gsTunnelInfo[i].nSID);
        }
    }
}

//########################################################
//# Main
//########################################################
int main(int argc, char *argv[]) {
    int ret = 0;
    const char *license_key = NULL;
    LogAttr log_attr = {0};
    char log_path[] = "./client_log.txt";
    char *args = strdup(__FILE__);

    if (argc < 2) {
        printf("%s <UID1> [UID2] [UID3] [UID4]\n", argv[0]);
        return -1;
    }

    for (int i = 0; i < MAX_SERVER_CONNECT_NUM; i++) {

        memset(gsTunnelInfo[i].uid, 0, sizeof(gsTunnelInfo[i].uid));
        gsTunnelInfo[i].nHttpIndex = -1;
        gsTunnelInfo[i].nHttpsIndex = -1;
        gsTunnelInfo[i].nSID = -1;
        gsTunnelInfo[i].nTelnetIndex = -1;
        gsTunnelInfo[i].nSshIndex = -1;
        gsTunnelInfo[i].nRetryConnectFailedCnt = 0;
        gsTunnelInfo[i].bException = false;

        if(i >= argc -1){
            continue;
        }

        if (strlen(argv[i + 1]) != 20) {
            printf("Invalid UID format\n");
            return -1;
        }

        strcpy(gsTunnelInfo[i].uid, argv[i + 1]);
        printf("Tunnel Server UID[%s]\n", gsTunnelInfo[i].uid);
    }

    //===== set license key for enable TUTK SDK =====
    license_key = getenv(ENV_LICENSE_KEY);
    if (license_key == NULL) {
        printf("Environment variable %s must be set!!\n", ENV_LICENSE_KEY);
        return -1;
    }

    ret = TUTK_SDK_Set_License_Key(license_key);
    printf("TUTK_SDK_Set_License_Key() ret[%d]\n", ret);
    if (ret < 0) {
        return -1;
    }

    //==== create P2PTunnel agent =====
    log_attr.path = log_path;
    ret = P2PTunnel_Set_Log_Attr(log_attr);
    printf("P2PTunnel_Set_Log_Attr() ret[%d]\n", ret);

    ret = P2PTunnelAgentInitialize2(1, 0);
    if (ret != TUNNEL_ER_NoERROR) {
        printf("P2PTunnelAgentInitialize2() ret[%d]\n", ret);
        return -1;
    }

    P2PTunnelAgent_GetStatus(tunnelStatusHandle, (void *)args);

    for (int i = 0; i < MAX_SERVER_CONNECT_NUM; i++) {
        if (strlen(gsTunnelInfo[i].uid) == 0) {
            continue;
        }
        tunnelAgentConnect(&gsTunnelInfo[i], i);
    }

    while (1) {
        for (int i = 0; i < MAX_SERVER_CONNECT_NUM; i++) {
            if (gsTunnelInfo[i].nSID >= 0) {
                int access_time = P2PTunnel_LastIOTime(gsTunnelInfo[i].nSID);
                if (access_time < 0) {
                    printf("P2PTunnel_LastIOTime() UID[%s] error[%d]\n", gsTunnelInfo[i].uid, access_time);
                } else {
                    printf("UID[%s]:%u, \n", gsTunnelInfo[i].uid, access_time);
                }
            }

            if (strlen(gsTunnelInfo[i].uid) != 0 && gsTunnelInfo[i].nSID < 0) {
                printf("Reconnect to UID[%s]\n", gsTunnelInfo[i].uid);
                tunnelAgentConnect(&gsTunnelInfo[i], i);

                if (gsTunnelInfo[i].nSID < 0) {
                    if (++gsTunnelInfo[i].nRetryConnectFailedCnt > 3) {
                        printf("Retry connection timeout UID[%s]\n", gsTunnelInfo[i].uid);
                        memset(gsTunnelInfo[i].uid, 0, sizeof(gsTunnelInfo[i].uid));
                    }
                }
            }
        }
        sleep(1);
        tunnelExceptionCheck();
    }

    tunnelAgentDisconnectAll();
    P2PTunnelAgentDeInitialize();
    return 0;
}
